/********************************************************************
/*Name:		Shuang Xu
/*DePaul ID:	0255454
/*Course No:	SE455, Software Development Methods
/*Instructor:	Robert Stalzer
/*Project Name: Tetris
/*******************************************************************/

/*
This file defines class PlayingField which is an extension of
Canvas component. Its responsibilities are:

(1) Listen to keyPressed event and carry out one
    of the following actions depending on a key:              
		a. Move figure to left or right
		b. Rotate the figure
		c. Drop the figure
 
(2) Track filled cells and rows. If a whole row 
    has been filled up, eliminate it.

(3) Check that the height of the pile does not 
	exceed the limit.

(4) Check when a figure has hit the bottom.
*/

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.*;


public class PlayingField 
    extends Field
{

    //********************** Variables **********************

    // Some useful constants

    // Points for landing a figure
    public static final int POINTS_LAND = 10;		
    // Points for filling up one row
    public static final int POINTS_PER_ROW = 500;  
    // Points for filling up four rows in a row
    public static final int POINTS_PER_FOUR_ROWS = 3000;  
    // Points for advancing to the upper lever
    public static final int POINTS_FOR_UPLEVEL = 5000;  
   
    // Speeds for each level in milliseconds. There are twelve levels.
    public final static int levelSpeeds[] = 
    { 1200, 1100, 1000, 900, 800, 700, 600, 500, 400, 300, 200, 100, 80};  

    // Rows and Columns to divide playing field into.
    // These should be set before the game starts by calling 
    //          setDimensions() method.
    // DEFAULT SETTINGS ARE: rows = 20, cols = 10
    protected int rows;
    protected int cols;

    // Width and height of each cell.
    protected int cellW;
    protected int cellH;    	
      
    // Array of cells holding boolean values. 
    // A value of true indicates that a cell is filled.
    protected boolean cells[][];

    // Array to hold the color of each cell to enable correct repainting
    protected Color cellColors[][];
    protected Color bkgColor;

    // Number of points scored
    protected int score;
    // Current level
    private int level; 

    // Current shape
    protected Shape sh = null;
    // Thread on which the game will be played
    protected Thread gameThread = null;       
    // KeyListener
    protected PlayingFieldKeyListener KeyListener = new PlayingFieldKeyListener();
    // MoveController
    protected MoveController mc = null;
    // Shape Factory is used to create shapes randomly
    protected ShapeFactory shapefactory = null;

    // Number of milliseconds to pause gameThread
    int milliseconds = 10;

    // Cheat code
    private String cheatCode;

    //********************** Methods **********************
    //
    //	Public constructor
    public PlayingField(int cols, int rows, Color color)
    {
        // color of the playingfield
	if (color != null) bkgColor = color;
	// Set default dimensions
        if (cols>0 && rows>0) setDimensions(rows, cols);
	// color of the canvas
	setBackground(Color.lightGray); 
        //decault cell size	
        cellW = 19;
        cellH = 19;

        addKeyListener(KeyListener);
    }

    // Set all cells to empty
    private void resetCells()
    {
        for ( int x = 0; x < rows; x++ ) { 
            for ( int y = 0; y < cols; y++ ) { 
                cells[x][y] = false;
                cellColors[x][y] = bkgColor;
            }
        }
    }

    // Returns true if at least one column in a playing field 
    // has been filled up
    public boolean gameOver()
    {
        int c;     // Current column
        for ( c = 0; c < cols; c++ ) 
            if ( cells[0][c] )    return true;
        return false;
    }

    // Returns true if shape cannot move down any further
    public boolean hitBottom( Shape s)
    {
        FilledCells f = s.getFilledCells();
        Cells l  = s.getLowestCell();

        // If the figure has hit the bottom of the playing field
        if ( l.row >= rows - 1 ) {
            fillCells( f );
            checkRowsForCompletion();
            return true;
        }

        // If the figure is on top of a filled cell
        for ( int x = 0; x < f.length; x++ ) {       
            if ( cells[f.cells[x][0] + 1][f.cells[x][1]] == true ) {
                fillCells( f );
                checkRowsForCompletion();
                return true;
            }
        }

        return false;	// Haven't hit the bottom yet
    }

    // Check if the figure is occupying an already filled cell
    public boolean isInvading( FilledCells fc )
    {
        // Check that the figure does not move into a filled cell
        for ( int r = 0; r < rows; r++ ) 
            for ( int c = 0; c < cols; c++ ) 
                if ( cells[r][c] == true ) 
                    for ( int x = 0; x < fc.length; x++ ) 
                        if ( fc.cells[x][0] == r && fc.cells[x][1] == c )
                            return true;  
        
        // Check that the figure does not go below last row
        for ( int a = 0; a < fc.length; a++ )
            if ( fc.cells[a][0] >= rows )
                return true;

        return false;
    }
        
    // Check if the figure is out of bounds horizontally
    public boolean outOfBounds( FilledCells fc ) 
    {
        for ( int x = 0; x < fc.length; x++ ) 
            if ( fc.cells[x][1] < 0 || fc.cells[x][1] >= cols )
                return true;
        return false;      
    }  

    //	Return width of each cell
    public int getCellWidth() 
    {
        return cellW;    
    }

	// Return height of each cell
    public int getCellHeight()
    {
        return cellH;
    }

    // Returns number of rows
    public int getNumberOfRows()
    {
        return rows;
    }

    // Returns number of columns
    public int getNumberOfColumns()
    {
        return cols;
    }

    // Returns KeyListener
    public KeyAdapter getKeyListener()
    {
        return KeyListener;
    }

    // Set how many rows and columns playing field will have
    public void setDimensions( int r, int c )
    {
        rows = r;
        cols = c;
        cells = new boolean[rows][cols];
        cellColors = new Color[rows][cols];
        resetCells();
    }
            
    // Get the size of the playingfield
    public Dimension getFieldSize()
    {
        return new Dimension( cellW*cols, cellH*rows );
    }
 
    // Randomly pick the shape
    public Shape pickShape()
    {
        String shapes[] = { "LShape", "BackwardLShape", "StairShape", "BackwardStairShape", "StickShape", "CubeShape", "TShape" };
        int numShapes = shapes.length;
        int pickedShape = (int)(Math.random() * numShapes);
        //((Applet)getParent()).showStatus( "Random number is: " + pickedShape );
            
	shapefactory = new ShapeFactory(pickedShape);
	return shapefactory.Build();
    }

    // Start the game
    public void startNewGame()
    {
        score = 0;
        level = 0;
        cheatCode = "";
        resetCells();	// All cells are empty initially
        Graphics g = getGraphics();
	g.setColor( bkgColor);
	int w = getFieldSize().width;
	int h = getFieldSize().height;
        g.fillRect( (getSize().width - w)/2, (getSize().height - h)/2, w-1, h );
        g.dispose();
        gameThread = null;
        gameThread = new Thread( this );
        gameThread.start();
    }
	
    public void suspendGame() {
	gameThread.suspend();
    }
    public void resumeGame() {
	gameThread.resume();
    }
    public void stopGame() {
	gameThread.stop();
    }

    // Game loop
    public void run()
    {
	while( !gameOver() ) {
            sh = null;
            sh = pickShape();               
            sh.setCoords( 0, cols/2 ); // drop from the mid-top
            sh.draw();
            milliseconds = levelSpeeds[level];
            while( !hitBottom( sh ) ) {
       	  	// Drop/Advance a row
       	  	mc = new DropController( sh );
       	  	mc.display();		
		// Wait before dropping one more level
                pauseGame( milliseconds );	
            }
            score += POINTS_LAND;
            checkLevel();
        }
    }          

    public void fillCells( FilledCells fc )
    {
        if ( fc == null )    return;
        for ( int x = 0; x < fc.length; x++ ) {
            cells[fc.cells[x][0]][fc.cells[x][1]] = true; 
            cellColors[fc.cells[x][0]][fc.cells[x][1]] = sh.getShapeColor();
        }
    }

    public void paint( Graphics g )
    {
        for ( int r = 0; r < rows; r++ ) {
            for ( int c = 0; c < cols; c++ ) {
                if ( cells[r][c] ) {
                    g.setColor( cellColors[r][c]  );
                    paintCell( g, r, c);
                }
            }
        }
    }

    public void paintCell( Graphics g, int cellRow, int cellCol)
    {
        if ( g != null )
		g.fill3DRect( cellCol * cellW, cellRow * cellH, cellW - 1, cellH - 1, true );
    }

    // Listens to keyboard events generated by the player 
    // and responds to them accordingly
    class PlayingFieldKeyListener extends KeyAdapter
    {	
        public void keyPressed( KeyEvent e )
        {   
            if ( sh == null || milliseconds == 10 ) {
                e.consume();
                return;
            }
	    int keyCode = e.getKeyCode();
            
            switch ( keyCode ) { 
         	case( 37 ): 		// Left key - move left
			mc = new LeftKeyController(sh);
			mc.display();
                     	break;
                case( 38 ):	 	// Up key - rotate
			mc = new RightKeyController(sh);
			mc.display();
                     	break;
                case( 39 ):		// Right key - move right
			mc = new UpKeyController(sh);
			mc.display();
                     	break;
                case( 40 ):	       // Down key
			milliseconds = milliseconds/2;  // Speed up
                     	break;
                default:
                    	cheatCode += e.getKeyChar();
                    	if ( cheatCode.equals( "lev6" ) )     score = 42000;
                    	if ( cheatCode.length() >= 4 )        cheatCode = "";
                    	break;
            }
        }
    }
    
    // Pause game thread for specified number of milliseconds
    public void pauseGame( int millis )
    { 
        try {
            gameThread.sleep( millis );
        }
        catch ( InterruptedException e ) {;}
    }

    // Returns the score
    public int getScore()
    {
        return score;
    }
    public Color getBackground() 
    {
	return bkgColor;
    }
 
    // Check if any rows have been completely filled and remove them
    private synchronized void checkRowsForCompletion()
    {
        int firstCompletedRow = - 1, lastCompletedRow = -1;

        for ( int r = 0; r < rows; r++ ) {
            boolean rowCompleted = true;	// Assume the current row is completed
            for ( int c = 0; c < cols; c++ ) {
                if ( cells[r][c] == false )    
                    rowCompleted = false;
            }
            if ( rowCompleted ) {
              if ( firstCompletedRow == -1 ) 
                    firstCompletedRow = lastCompletedRow = r;
 		else
                    lastCompletedRow = r;
		    }
        }
        if ( firstCompletedRow != - 1 ) {
            removeRows( firstCompletedRow, lastCompletedRow ); 

            int completedRows = lastCompletedRow - firstCompletedRow + 1;

            if ( completedRows < 4 )      score += (completedRows * POINTS_PER_ROW);
            else if ( completedRows == 4 )  score += POINTS_PER_FOUR_ROWS;
        }
    }

    // Remove rows specified by startRow and endRow 
    private void removeRows( int startRow, int endRow )
    {
        for ( int r = startRow; r <= endRow; r++ ) {

	    // Empty this row
            for ( int c = 0; c < cols; c++ ) {
                cells[r][c] = false;
            }
                    
	    // Move all rows on top one level down
            for ( int x = r - 1; x >= 0; x-- ) {
                for ( int y = 0; y < cols; y++ ) {
                    cells[x + 1][y] = cells[x][y];
                    cellColors[x + 1][y] = new Color( cellColors[x][y].getRGB() );
                    cells[x][y] = false;
                }
            }
        }
        repaint();	// Paint new layout
    }
               
    //  Method that checks if it's time to advance to the next level
    private void checkLevel() 
    {
        int pts = POINTS_FOR_UPLEVEL;	// Points needed to advance to next level

        for ( int x = 12; x >= 0; x-- ) {
            if ( score >= x * pts ) {   
                level = x;
                return;
            }
        }               
    }

    // Set level
    public void setLevel( int l )
    {
        level = (l >= 0 && l < 12) ? l : level;
    }

    // Return current level
    public int getLevel()
    {
        return level;
    }

}      
